/*
	Simple VGA PEL Shift test v0.06

	(c) 2012 Angelo Salese and the MESS team

	TODO:
	- Add a manual/stop mode?
	- getch() is annoying, keys gets stuck in buffer.
	- test #2 black screen time is too much. I either need to change
	  the VGA pixel writes (imho it's too slow) or print a message
	  on screen that says "now loading".

	Version log:
	v0.06
	- Added a new item test. Tests out bit 0 of the PEL shift via a
	  simple color cycle algorythm.
	- Made various clean-ups.

	v0.05
	- Added a necessary out write to MovePELshift, if you don't
	  write bit 5 after writing to the attribute register it won't
	  work on real HW (i.e. system doesn't know what are the palette
	  entries).

	v0.04
	- Added disable/enable irqs in MovePELshift sub-routine

	v0.03
	- defined a bunch of palette entries
	- fixed a typo
	- some code clean-ups

	v0.02
	- Made the timer to be slower
	- Written a custom charset, and wrote a drawing routine for that

	v0.01
	- Initial Release
*/

#include <stdio.h>
#include <conio.h>
#include <dos.h>

unsigned char get_video_mode;
unsigned char cur_PELshift;
int PELtimer;
unsigned char read_input;
unsigned char exit_routine;

void InitVGA256();
void InitVRAMN1();
void InitVRAMN2();
void InitVGAPalette();
void ClearVGAPalette();
void WritePal(char index, char r, char g, char b);
void InitTextMode();
void TestN1();
void TestN2();
void MovePELShift(char param);
void PrintChar(char param,int x,int y);

void main()
{
	clrscr();
	textcolor(15);
	gotoxy(24,1); cprintf("Simple VGA PEL Shift test v0.06");
	gotoxy(1,3); cprintf("(c) 2012 Angelo Salese and the MESS team");
	gotoxy(1,4); cprintf("Press any key to start, any key to move on after each test item");
	getch();
	InitVGA256();
	InitVRAMN1();
	TestN1();
	InitVRAMN2();
	TestN2();
	InitTextMode();
	gotoxy(1,1); cprintf("Test ended, press any key to return to TC or DOS");
	getch();
}

void TestN1()
{
	cur_PELshift = 0;
	PELtimer = 0;
	exit_routine = 0;
	while(!exit_routine)
	{
		/* if input is pressed go outside of this function */
		if(kbhit())
		{
			exit_routine = 1;
			getch(); // dummy
		}
		/* Check VBlank, and change PEL shift accordingly */
		MovePELShift(cur_PELshift);
		PELtimer++;

		if(PELtimer > 70)/* change PELtimer every 70 frames (more or less 1 sec) */
		{
			cur_PELshift++;
			cur_PELshift&=0xf;
			PELtimer = 0;
		}

		/* Now print the chars on screen */
		PrintChar(cur_PELshift,32,16);
		PrintChar(PELtimer / 10,8,16);
		PrintChar(PELtimer % 10,16,16);
	}
}

void TestN2()
{
	cur_PELshift = 0;
	PELtimer = 0;
	exit_routine = 0;
	while(!exit_routine)
	{
		/* if input is pressed go outside of this function */
		if(kbhit())
		{
			exit_routine = 1;
			getch(); // dummy
		}
		/* Check VBlank, and change PEL shift accordingly */
		MovePELShift(cur_PELshift);
		PELtimer++;

		if(PELtimer > 70)/* change PELtimer every 70 frames (more or less 1 sec) */
		{
			cur_PELshift++;
			cur_PELshift&=0x1;
			PELtimer = 0;
		}

		/* Now print the chars on screen */
		PrintChar(cur_PELshift,32,16);
		PrintChar(PELtimer / 10,8,16);
		PrintChar(PELtimer % 10,16,16);
	}
}


/* set mode 13h */
void InitVGA256()
{
	asm{
		/* get video mode, for resuming it later on */
		mov ah,0x0f
		int 0x10
		mov get_video_mode,al
		/* set mode 0x13 */
		mov ah,0x00
		mov al,0x13
		int 0x10
	}
}

/* setup palette */
void InitVGAPalette()
{
	WritePal(0x10,0x04,0x04,0x04);
	WritePal(0x11,0x3f,0x3f,0x3f);
	WritePal(0x12,0x3f,0x2f,0x1f);

	WritePal(0x24,0x1f,0x1f,0x3f);
	WritePal(0x42,0x00,0x00,0x1f);
	WritePal(0xff,0x3f,0x14,0x07);
}

void ClearVGAPalette()
{
	int i;

	for(i=0x10;i<0x100;i++)
		WritePal(i,0x00,0x00,0x00);
}

void WritePal(char index, char r, char g, char b)
{
	asm{
		mov dx,0x3c8
		mov al,index
		out dx,al
		inc dx
		mov al,r
		out dx,al
		mov al,g
		out dx,al
		mov al,b
		out dx,al
	}
}

/* Return to whatever Video Mode was before */
void InitTextMode()
{
	asm{
		mov al,get_video_mode
		mov ah,0x00
		int 0x10
	}
}

/* Draw a vertical line at the center of the buffer */
void InitVRAMN1()
{
	int y;

	ClearVGAPalette();

	for(y=0;y<200;y++)
	{
		asm{
			mov ah,0x0c //plot pixel
			mov al,0x12 //color entry
			mov cx,320/2 // x
			mov dx,y     // y
			int 0x10
		}
	}

	InitVGAPalette();
}

void InitVRAMN2()
{
	int x,y;
	char col;

	ClearVGAPalette();

	for(x=0;x<320;x++)
	{
		for(y=0;y<200;y++)
		{
			if((x % 32) == 16)
				col = 0xff;
			else if(x & 0x10)
				col = 0x42;
			else
				col = 0x24;

			asm{
				mov ah,0x0c //plot pixel function
				mov al,col //color entry
				mov cx,x // x
				mov dx,y     // y
				int 0x10
			}
		}
	}

	InitVGAPalette();
}

void MovePELShift(char param)
{
	/* disable irqs */
	disable();
	asm{
		mov dx,0x3da
	}
// just reset the attribute state and check vblank while at it
vblank:
	asm{
		in al,dx
		test al,8
		jne vblank
	}
novblank:
	asm{
		in al,dx
		test al,8
		je novblank
	}
	asm{
		/* write to the PEL shift*/
		mov al,0x33 // index 0x13 + bit 5 for giving back palette control)
		mov dx,0x3c0
		out dx,al
		mov al,param
		out dx,al
	}
	/* enable irqs */
	enable();
}

/* custom 8x16 charset
 * Yeah, I know ... I could just check inside the VGA char
 * RAM buffer (one of the x86 irq vectors points to that)
 * and do some coding instead ... but for once I don't feel like
 * doing boring BCD conversions ;p
 */
static const char tile[16*0x10] =
{
	0x00, 0x7c, 0x82, 0x86, 0x8a, 0x8a, 0x92, 0x92, 0xa2, 0xa2, 0xc2, 0xc2, 0x82, 0x82, 0x7c, 0x00, // 0
	0x00, 0x08, 0x08, 0x18, 0x18, 0x28, 0x28, 0x48,	0x48, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00, // 1
	0x00, 0x3c, 0x42, 0x42, 0x04, 0x04, 0x08, 0x08,	0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x7e, 0x00, // 2
	0x00, 0x3c, 0x42, 0x02, 0x02, 0x02, 0x02, 0x1c,	0x02, 0x02, 0x02, 0x02, 0x02, 0x42, 0x3c, 0x00, // 3
	0x00, 0x08, 0x08, 0x18, 0x18, 0x18, 0x28, 0x28, 0x28, 0x48, 0x48, 0x7e, 0x08, 0x08, 0x08, 0x00, // 4
	0x00, 0x7e, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7c,	0x06, 0x02, 0x02, 0x02, 0x02, 0x06, 0x7c, 0x00, // 5
	0x00, 0x3c, 0x42, 0x40, 0x40, 0x40, 0x40, 0x7c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, // 6
	0x00, 0x7e, 0x02, 0x06, 0x04, 0x04, 0x0c, 0x08, 0x08, 0x18, 0x10, 0x10, 0x30, 0x20, 0x20, 0x00, // 7
	0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00, // 8
	0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3e, 0x02, 0x02, 0x02, 0x02, 0x02, 0x42, 0x3c, 0x00, // 9
	0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x00, // A
	0x00, 0x7c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7c,	0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7c, 0x00, // B
	0x00, 0x3c, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x3c, 0x00, // C
	0x00, 0x7c, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7c, 0x00, // D
	0x00, 0x7e, 0x40, 0x40, 0x40, 0x40, 0x40, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7e, 0x00, // E
	0x00, 0x7e, 0x40, 0x40, 0x40, 0x40, 0x40, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x00  // F

};

void PrintChar(char param,int x,int y)
{
	int xi, yi;
	char cur_pix;

//	cprintf("%02x",param);

	for(yi=0;yi<16;yi++)
	{
		for(xi=0;xi<8;xi++)
		{
			cur_pix = (tile[param*16+yi] >> 7-xi) & 1;
			cur_pix += 0x10;
			asm{
				mov al,cur_pix
				mov ah,0x0c //plot pixel
				//mov al,0x0f // color
				mov bx,x
				add bx,xi
				mov cx,bx
				mov bx,y
				mov bx,yi
				mov dx,bx
				int 0x10
			}
		}
	}

}